import os

from delphi4py import DelPhi4py
from pdbmender.formats import gro2pdb, get_grobox_size, get_chains_from_file
from pdbmender.utils import identify_tit_sites

from pypka.log import checkDelPhiErrors
from pypka.clean.checksites import (
    check_sites_integrity,
    make_delphi_inputfile,
    fix_fixed_sites,
)
from pypka.clean.cleaning import cleanPDB, inputPDBCheck
from pypka.clean.pdb_out import write_output_structure
from pypka.concurrency import (
    runDelPhiSims,
    runInteractionCalcs,
    startPoolProcesses,
)
from pypka.config import Config
from pypka.constants import (
    KBOLTZ,
    TITRABLETAUTOMERS,
    TERMINAL_OFFSET,
    INSERTIONCODE_OFFSET,
)
from pypka.molecule import Molecule
from pypka.mc.run_mc import MonteCarlo
from pdbmender.utils import correct_insertion_codes

import logging

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
formatter = logging.Formatter("%(levelname)s: %(message)s")
ch.setFormatter(formatter)
logger.addHandler(ch)


class Titration:
    """
    Main PypKa class

    Attributes:
        params (Config): method parameters
        molecules (dict): titrating molecules ordered by chain
    """

    def __init__(
        self, parameters, sites="all", fixed_sites=None, debug=False, run="all"
    ):
        """
        Runs the pKa prediction

        Args:
            parameters (dict): input parameters to overwrite the default configuration.
            sites (dict, optional): tuple titrable residue numbers indexed by chain.
                                    Defaults to 'all' which will titrate all titrable residues.
            debug (boolean, optional): debug mode switch. Defaults to False.
        """
        self.molecules = {}
        self.__parameters = Config.storeParams(self, debug, parameters, sites)
        self.pKas = {}
        self.isoelectric_point = None
        self.isoelectric_point_limit = None

        print("Start Preprocessing")
        self.preprocessing(sites, fixed_sites)
        self.processDelPhiParams()

        if run == "preprocess":
            return

        print("Start PB Calculations")
        self.DelPhiLaunch()

        if run == "PB":
            return

        print("Calculating site-site interactions")
        self.calcSiteInteractionsParallel()

        print("Start MC", end="\r")
        self.run_mc()

        if Config.pypka_params["f_structure_out"]:
            sites = self.get_all_sites(get_list=True)
            write_output_structure(sites, self.molecules, self.delphi_input_content)

        print("Results")
        print(self)

        print("API exited successfully")

    def preprocessing(self, sites, fixed_sites):
        def create_tit_sites(fname, chains_res):
            for _, molecule in self.molecules.items():
                molecule.deleteAllSites()
            check_sites_integrity(fname, self.molecules, chains_res)

        if fixed_sites:
            for chain, fixed_sites_info in fixed_sites.items():
                if chain not in sites:
                    sites[chain] = []
                for sitenumb in fixed_sites_info.keys():
                    sites[chain].append(sitenumb)

        f_in = "input.pdb"
        if Config.pypka_params["f_in_extension"] == "gro":
            groname = Config.pypka_params["f_in"]
            gro2pdb(groname, f_in)
            box = get_grobox_size(groname)
            Config.pypka_params.setBox(box)
            Config.pypka_params.redefine_f_in(f_in)

        renumbered = correct_insertion_codes(
            Config.pypka_params["f_in"],
            f_in,
            insertion_offset=INSERTIONCODE_OFFSET,
            replace_empty_chain=False,
        )

        if not Config.pypka_params["f_in_extension"] == "pdb":
            f_format = Config.pypka_params["f_in_extension"]
            raise Exception(
                "{0} file format is not currently supported".format(f_format)
            )
        automatic_sites = self.create_molecule(sites, renumbered)

        # f_in = Config.pypka_params["f_in"]
        if not automatic_sites:
            # Reading .st files
            # If the titrable residues are defined
            clean_pdb = Config.pypka_params["clean_pdb"]
            _, chains_res = inputPDBCheck(f_in, sites, clean_pdb)
            for chain, molecule in self.molecules.items():
                molecule.loadSites(chains_res[chain])

        else:
            # If the titrable residues are not defined
            chains_res = identify_tit_sites(
                f_in,
                self.molecules.keys(),
                nomenclature=Config.pypka_params["ffinput"],
                add_ser_thr=Config.pypka_params["ser_thr_titration"],
            )
            self.assign_residues_to_molecule(chains_res)

        if Config.pypka_params["clean_pdb"]:
            # Creates a .pdb input for DelPhi
            # where residues are in their standard state
            f_out = "TMP.pdb"
            chains_res = cleanPDB(
                f_in, f_out, self.molecules, chains_res, automatic_sites
            )
            if not Config.debug and os.path.isfile(f_in):
                os.remove(f_in)
            f_in = f_out

        create_tit_sites(f_in, chains_res)

        f_out = "delphi_in_stmod.pdb"
        self.sequence = make_delphi_inputfile(f_in, f_out, self.molecules, fixed_sites)

        if fixed_sites:
            fix_fixed_sites(self.molecules, fixed_sites, f_out)

        if not Config.debug and os.path.isfile(f_in):
            os.remove(f_in)

    def create_molecule(self, sites, renumbered):
        # Creating instance of TitratingMolecule
        automatic_sites = False
        if sites == "all":
            f_in = Config.pypka_params["f_in"]
            chains = get_chains_from_file(f_in)
            sites = {chain: ["all"] for chain in chains}

        for chain, site_list in sites.items():
            if site_list == ["all"]:
                site_list = "all"
            else:
                site_list = [str(site) for site in site_list]
                sites[chain] = site_list
            renumbered_chain = renumbered[chain] if chain in renumbered.keys() else {}
            self.molecules[chain] = Molecule(chain, site_list, renumbered_chain)
            if site_list == "all":
                automatic_sites = True
        return automatic_sites

    def assign_residues_to_molecule(self, chains_res):
        if not chains_res:
            f_in = Config.pypka_params["f_in"]
            raise Exception("Not one titrable residue was found in {}".format(f_in))

        for chain, residues in chains_res.items():
            molecule = self.molecules[chain]
            for resnumb, resname in residues.items():
                if resname in ("NTR", "CTR"):
                    resnumb = int(resnumb)
                    if resname == "NTR":
                        molecule.NTR += [resnumb]
                    elif resname == "CTR":
                        molecule.CTR += [resnumb]
                    resnumb += TERMINAL_OFFSET

                sID = molecule.addSite(resnumb)
                ntautomers = TITRABLETAUTOMERS[resname]
                molecule.addTautomers(sID, ntautomers, resname)

        for molecule in self.molecules.values():
            # Adding the reference tautomer to each site
            molecule.addReferenceTautomers()
            # Assigning a charge set to each tautomer
            molecule.addTautomersChargeSets()

    def get_total_atoms(self):
        total_atoms = 0
        for molecule in self.molecules.values():
            total_atoms += molecule.getNAtoms()
        return total_atoms

    def get_all_sites(self, get_list=False):
        sites = {}
        if get_list:
            sites = []
        for chain, molecule in self.molecules.items():
            chain_sites = molecule.getSitesOrdered()
            if get_list:
                # if isinstance(chain_sites, dict):
                #    chain_sites = chain_sites.values()
                sites += chain_sites
            else:
                sites[chain] = chain_sites
        return sites

    def get_molecule_sites(self):
        sites = {}
        for chain, molecule in self.molecules.items():
            chain_sites = molecule.getSites()
            sites[chain] = chain_sites
        return sites

    def get_total_sites(self):
        total = 0
        for sites in self.get_all_sites().values():
            total += len(sites)
        return total

    def processDelPhiParams(self):
        # Storing DelPhi parameters and Creates DelPhi data structures
        logfile = "LOG_readFiles"

        delphimol = DelPhi4py(
            Config.pypka_params["f_crg"],
            Config.pypka_params["f_siz"],
            "delphi_in_stmod.pdb",
            igrid=Config.delphi_params["gsize"],
            scale=Config.delphi_params["scaleM"],
            precision=Config.delphi_params["precision"],
            epsin=Config.delphi_params["epsin"],
            epsout=Config.delphi_params["epssol"],
            conc=Config.delphi_params["ionicstr"],
            ibctyp=Config.delphi_params["bndcon"],
            res2=Config.delphi_params["maxc"],
            nlit=Config.delphi_params["nlit"],
            nonit=Config.delphi_params["nonit"],
            relfac=0.0,
            relpar=0.0,
            pbx=Config.delphi_params["pbx"],
            pby=Config.delphi_params["pby"],
            isurftype=Config.delphi_params["nanoshaper"],
            debug=Config.debug,
            outputfile=logfile,
        )

        if Config.pypka_params["f_structure_out"]:
            with open("delphi_in_stmod.pdb") as f:
                self.delphi_input_content = f.readlines()
        if Config.pypka_params["save_pdb"]:
            with open("delphi_in_stmod.pdb") as f, open(
                Config.pypka_params["save_pdb"], "w"
            ) as f_new:
                content = f.read()
                f_new.write(content)

        if not Config.debug:
            os.remove("delphi_in_stmod.pdb")

        checkDelPhiErrors(logfile, "readFiles")

        lookup_atoms = {}
        for molecule in Config.titration.molecules.values():
            for atom_name, atom_id, atom_position in molecule.iterAtoms():
                lookup_atoms[atom_position] = (atom_id, atom_name)

        # Stores DelPhi4py object and attributes
        Config.delphi_params.store_run_params(delphimol)
        Config.delphi_params.lookup_atoms = lookup_atoms

        if Config.debug:
            for chain, molecule in self.molecules.items():
                print("### CHAIN {chain} ###".format(chain=chain))
                molecule.printAllSites()
                molecule.printAllTautomers()
            print(delphimol)

    def iterAllSitesTautomers(self):
        """
        Generator that iterates through all Tautomer instances.

        The iteration is sorted by site and within each site the first
        to be yielded is the reference tautomer and then the rest of
        the tautomers by order
        """
        for molecule in self.molecules.values():
            for site in molecule.sites_order:
                yield site.ref_tautomer
                for tautomer in sorted(site.tautomers.values()):
                    yield tautomer

    def calcpKint(self, unpacked_results):
        """Calculation the pKint of all tautomers."""
        i = -1
        if Config.debug:
            print("############ results ############")
        pkints = ""
        contributions = ""
        for tautomer in self.iterAllSitesTautomers():
            i += 1
            core_index = i % Config.pypka_params["ncpus"]
            job_index = int(i / Config.pypka_params["ncpus"])
            result = unpacked_results[core_index][job_index]

            tautname = result[0]
            # tautresnumb = result[1]

            esolvationM = result[2]
            sitpotM = result[3]
            esolvationS = result[4]
            sitpotS = result[5]

            tautomer = Config.parallel_params.all_tautomers_order[i]

            tautomer.saveDelPhiResults(esolvationS, sitpotS, esolvationM, sitpotM)
            if Config.debug:
                print("### new tautomer ###")
                print(
                    (
                        i,
                        core_index,
                        job_index,
                        tautname,
                        tautomer.name,
                        esolvationM,
                        esolvationS,
                    )
                )
                print(
                    (
                        tautomer.name,
                        tautomer.esolvationS,
                        len(tautomer.sitpotS),
                        tautomer.esolvationM,
                        len(tautomer.sitpotM),
                    )
                )

            tautomer.calcBackEnergy()
            if not tautomer.isRefTautomer():
                tautomer.calcpKint()
                if Config.debug:
                    print(("pkint", tautomer.name, tautomer.dg, id(tautomer)))
                    pkints += "{} {} {}\n".format(
                        tautomer.name, tautomer.dg, tautomer.pKint
                    )
                    contributions += "{}{} {} {} {} {}\n".format(
                        tautomer.site.res_number,
                        tautomer.name,
                        tautomer.dG_solvationM,
                        tautomer.dG_solvationS,
                        tautomer.dG_solvationM - tautomer.dG_solvationS,
                        tautomer.dG_back,
                    )
        if Config.debug:
            with open("pkint", "w") as f_new1, open("contributions", "w") as f_new2:
                f_new1.write(pkints)
                f_new2.write(contributions)

    def calcSiteInteractionsParallel(self):
        """
        Calculates the pairwise interaction energies

        Interactions are calculated using a pool of processes
        and written in a formatted .dat file

        Args:
          ncpus (int): number of cpus to be used
        """

        def writeDatHeader(sites):
            """Writes pKint energies in .dat file header."""
            to_write = "{0}\n".format(len(sites))
            for site in sites:
                to_write += "{0:3s}-{1:<7}{2:>2}  P  *\n".format(
                    site.res_name, site.res_number, len(site.tautomers) + 1
                )
                if site.type == "c":
                    tau_prot_state = 0
                    ref_prot_state = 1
                elif site.type == "a":
                    tau_prot_state = 1
                    ref_prot_state = 0

                for tautomer in site.iterOrderedTautomersWithoutRef():
                    to_write += "{0:1d} {1:13.6e}\n".format(tau_prot_state, tautomer.dg)
                to_write += "{0:1d}  0.000000e+00\n".format(ref_prot_state)
            if Config.debug:
                with open("interactions.dat", "w") as f_new:
                    f_new.write(to_write)

        Config.loadParams(self.__parameters)

        sites = self.get_all_sites(get_list=True)
        if Config.debug:
            writeDatHeader(sites)

        counter = 0
        site_interactions = []
        for site1 in sites[:-1]:
            counter += 1
            for site2 in sites[counter:]:
                site_interactions.append((site1, site2))

        npossible_states = [len(site.tautomers) + 1 for site in sites]
        Config.parallel_params.npossible_states = npossible_states

        interactions = []
        for nstate1 in range(sum(npossible_states)):
            interactions.append([])
            for _ in range(sum(npossible_states)):
                interactions[nstate1].append(-999999)

        interactions_look = []
        aux = -1
        site = -1
        for nstates in npossible_states:
            site += 1
            interactions_look.append([])
            for _ in range(nstates):
                aux += 1
                interactions_look[site].append(aux)

        Config.parallel_params.interactions_look = interactions_look
        Config.parallel_params.site_interactions = site_interactions
        Config.parallel_params.all_sites = sites
        Config.parallel_params.interactions = interactions
        ncpus = min(len(site_interactions), Config.pypka_params["ncpus"])
        results = []

        if ncpus > 0:
            results = startPoolProcesses(
                runInteractionCalcs,
                site_interactions,
                ncpus,
                assign="ordered",
                merged_results=True,
            )

        interactions = Config.parallel_params.interactions
        to_write = ""
        temperature = float(Config.pypka_params["temp"])
        for interaction in results:
            site1i = interaction[0]
            site2i = interaction[1]
            if interactions[site1i][site2i] != -999999:
                interactions[site1i][site2i] += interaction[2]
                interactions[site2i][site1i] += interaction[2]
                interactions[site1i][site2i] /= 2
                interactions[site2i][site1i] /= 2

                if Config.debug:
                    col1 = interaction[3][6:18]
                    col2 = interaction[3][:6]
                    col3 = interactions[site1i][site2i] * (KBOLTZ * temperature)
                    to_write += "{0}{1} {2:13.6e}\n".format(col1, col2, col3)
            else:
                interactions[site1i][site2i] = interaction[2]
                interactions[site2i][site1i] = interaction[2]

        if Config.debug:
            with open("interactions.dat", "a") as f_new:
                f_new.write(to_write)

    def DelPhiLaunch(self):
        Config.loadParams(self.__parameters)

        if self.get_total_sites() < 1:
            raise Exception("At least one site has to be correctly defined.")

        all_tautomers = list(self.iterAllSitesTautomers())

        Config.parallel_params.all_tautomers_order = all_tautomers

        # Runs DelPhi simulations for all tautomers
        results = startPoolProcesses(
            runDelPhiSims, all_tautomers, Config.pypka_params["ncpus"]
        )

        print("\rPB Runs Ended{:>80}".format(""))

        # Calculates the pKint of all tautomers
        print("Calculating intrinsic pK values")
        self.calcpKint(results)

    def run_mc(self):
        Config.loadParams(self.__parameters)

        sites = self.get_all_sites(get_list=True)

        mc = MonteCarlo(sites)
        text_pks, self.tit_curve = (mc.text_pks, mc.total_tit_curve)

        print("\rMC Runs Ended{:>80}\n".format(""))

        if Config.pypka_params["isoelectric_point"]:
            self.getIsoelectricPoint()
            if self.isoelectric_point:
                text_pks += "\nPredicted Isoelectric Point: {0}\n".format(
                    self.isoelectric_point
                )
            elif self.isoelectric_point_limit:
                pH, limit, charge = self.isoelectric_point_limit
                text_pks += "\nPredicted Isoelectric Point: {0} Limit: {1} \nCharge at pH {0}: {2}\n".format(
                    pH, limit, charge
                )

        if Config.pypka_params["f_out"]:
            with open(Config.pypka_params["f_out"], "w") as f:
                f.write(text_pks)
        if Config.pypka_params["f_prot_out"]:
            with open(Config.pypka_params["f_prot_out"], "w") as f:
                f.write(mc.text_prots)
        if Config.pypka_params["coupled_sites_output"]:
            with open(Config.pypka_params["coupled_sites_output"], "w") as f:
                f.write(mc.text_coupled_sites)

    def getTotalProtonationCurve(self):
        return self.tit_curve

    def getTitrationCurve(self):
        if Config.pypka_params["sites"] != "all":
            warn_message = "The titration curve weighted by charges can only be calculated when titrating all sites."
            raise Exception(warn_message)

        sites = self.get_all_sites(get_list=True)
        nsites = len(sites)
        total_anionic_titres = 0
        for site in sites:
            if site.type == "a":
                total_anionic_titres += 1

        total_arg_charge = 0
        for _, aas in self.sequence.items():
            for _, resname in aas.items():
                if resname == "ARG":
                    total_arg_charge += 1

        titration_curve = {}
        for pH, prot in self.tit_curve.items():
            prot = prot * nsites
            charge = prot - total_anionic_titres + total_arg_charge
            titration_curve[pH] = charge
        return titration_curve

    def getIsoelectricPoint(self):
        def calc_isoelectric_point():
            # anionic_aas = ("ASP", "GLU", "CYS", "TYR", "CTR")
            # cationic_aas = ("LYS", "ARG", "HIS", "NTR")

            titration_curve = [
                (pH, prot) for pH, prot in self.getTitrationCurve().items()
            ]
            i = 0
            pH, prot = titration_curve[i]
            while prot > 0.0 and i < len(titration_curve) - 1:
                i += 1
                pH, prot = titration_curve[i]

            if i == 0 or i == len(titration_curve) - 1:
                charge = round(prot, 2)
                warn_message = (
                    "Not enough points to interpolate the Isoelectric Point\n"
                    "Lowest pH value: {0}\t"
                    "Charge: {1}".format(pH, charge)
                )
                print(warn_message)
                limit = "-" if i == 0 else "+"
                return pH, limit, charge
            else:
                last_pH, last_prot = titration_curve[i - 1]

                isoelectric_point = pH - (prot * (last_pH - pH)) / (last_prot - prot)
                return isoelectric_point

        if not self.isoelectric_point:
            result = calc_isoelectric_point()
            if isinstance(result, tuple):
                self.isoelectric_point_limit = result
            else:
                self.isoelectric_point = result
        else:
            result = self.isoelectric_point

        return result

    def __iter__(self):
        self.itersites = self.get_all_sites(get_list=True)
        self.iternumb = -1
        self.itermax = len(self.itersites)
        return self

    def __next__(self):
        self.iternumb += 1
        if self.iternumb < self.itermax:
            site = self.itersites[self.iternumb]
            return site
        else:
            raise StopIteration

    @staticmethod
    def getParameters():
        """Get the parameters used in the calculations."""
        return "{}\n{}\n{}".format(
            Config.pypka_params.__str__(),
            Config.delphi_params.__str__(),
            Config.mc_params.__str__(),
        )

    @staticmethod
    def getParametersDict():
        return (
            Config.pypka_params.get_clean_params(),
            Config.delphi_params.get_clean_params(),
            Config.mc_params.get_clean_params(),
        )

    def getSiteInteractions(self):
        Config.loadParams(self.__parameters)

        return (
            Config.parallel_params.all_sites,
            Config.parallel_params.npossible_states,
            Config.parallel_params.interactions_look,
            Config.parallel_params.interactions,
        )

    def __getitem__(self, chain):
        return self.molecules[chain]

    def __str__(self, str_format="print"):
        if str_format == "print":
            output = "Chain  Site   Name      pK"
        elif str_format == "file":
            output = ""
        sites = self.get_all_sites()

        for chain in sites.keys():
            for site in sites[chain]:
                pk = site.pK
                if pk and str_format == "print":
                    pk = "{:.2f}".format(round(pk, 2))
                elif not pk:
                    pk = "Not In Range"
                output += "\n{0:>4} {1:>6}    {2:3}    {3:>5}".format(
                    chain, site.getResNumber(correct_icode=True), site.res_name, pk
                )

        if self.isoelectric_point:
            output += "\n\nPredicted Isoelectric Point: {0}\n".format(
                self.isoelectric_point
            )

        return output


def getTitrableSites(pdb, ser_thr_titration=True, debug=False):
    """
    Gets the all titrable sites from the pdb

    Returns an input structure to the Titration class containing all
    titrable sites found in the pdb file.

    Args:
        pdb (str): The filename a PDB file

    Returns:
        A dict mapping all titrable sites found in the pdb file
    """
    parameters = {
        "structure": pdb,
        "ncpus": 1,
        "epsin": 15,
        "ser_thr_titration": ser_thr_titration,
    }
    Config.storeParams("", debug, parameters, "all")

    chains = get_chains_from_file(pdb)
    sites = {chain: "all" for chain in chains}

    molecules = {}
    for chain, site_list in sites.items():
        molecules[chain] = Molecule(chain, site_list, [])

    chains_res = identify_tit_sites(pdb, chains)

    out_sites = {chain: [] for chain in chains_res.keys()}

    for chain, sites in chains_res.items():
        for resnumb, resname in sites.items():
            out_site = resnumb
            if resname == "NTR":
                out_site += "N"
            elif resname == "CTR":
                out_site += "C"

            out_sites[chain].append(out_site)

    return out_sites, chains_res



### USER INPUT ##

rt = 2  # Options: 2 5 'zero' 'min'

out_siz = "tmp.siz"
out_crg = "tmp.crg"

###

water_atype = "OW"

f_nonbonded = "ffG54a7pHtnb.itp"
f_rtp = "ffG54a7pHt.rtp"
old_siz_db = "DataBaseT_template.siz"
old_crg_db = "DataBaseT_template.crg"  # NTR and CTR charges are copied from here

ignore_list = ()

### MAIN ###

# when adding new titratable residues please include them here
sizes = {
    "HI2": ("HI0", "HI1"),
    "AS4": ("AS0", "AS1", "AS2", "AS3"),
    "LY3": ("LY0", "LY1", "LY2"),
    "GL4": ("GL0", "GL1", "GL2", "GL3"),
    "TY2": ("TY0", "TY1"),
    "CY2": ("CY0", "CY1"),
    "GD4": ("GD0", "GD1", "GD2", "GD3"),
    "PA4": ("PA0", "PA1", "PA2", "PA3"),
    "AR4": ("AR0", "AR1", "AR2", "AR3"),
    "PJ2": ("PJ0", "PJ1"),
}

rules = {
    "NTR": {
        "N": "nl",
        "H1": "h",
        "H2": "h",
        "H3": "h",
        "CA": "ch1",
        "C": "c",
        "O": "o",
    },
    "CTR": {
        "C": "c",
        "O1": "om",
        "O2": "om",
        "HO11": "h",
        "HO12": "h",
        "HO21": "h",
        "HO22": "h",
    },
}

rules["NT0"] = rules["NTR"]
rules["NT1"] = rules["NTR"]
rules["NT2"] = rules["NTR"]
rules["NT3"] = rules["NTR"]
rules["CT0"] = rules["CTR"]
rules["CT1"] = rules["CTR"]
rules["CT2"] = rules["CTR"]
rules["CT3"] = rules["CTR"]
rules["CT4"] = rules["CTR"]


def radius_from_lennard(c6, c12, rt="min", water=False):
    def calc_radius(rmin, c6, c12, rt):
        if isinstance(rt, int):
            kT = 2.49432
            radius = (rmin ** (-6) + (rt * kT / c12) ** 0.5) ** (-1 / 6)
        elif rt == "zero":
            radius = rmin * 2 ** (-1 / 6)
        else:
            radius = rmin
        return radius

    if not c6:
        return 0.001

    rmin = (2 * c12 / c6) ** (1.0 / 6)

    radius = calc_radius(rmin, c6, c12, rt)

    if water:
        return 10 * radius / 2
    else:
        return 10 * radius - water_radius


db_lj = {}
db_lj_atomic = {}
with open(f_nonbonded) as f:
    trigger = False
    for line in f:
        if ";" in line:
            comment_i = line.index(";")
            line = line[:comment_i]
        cols = line.strip().split()
        if not line.startswith(";") and "nonbond_params" in line:
            trigger = True
        elif "[ " in line:
            trigger = False

        if trigger and len(cols) == 5:
            if cols[0] == water_atype:
                atype = cols[1].lower()
            elif cols[1] == water_atype:
                atype = cols[0].lower()
            else:
                continue

            c6, c12 = float(cols[-2]), float(cols[-1])
            db_lj[atype] = (c6, c12)

db_res = {}
db_charges = {}
trigger = False
with open(f_rtp) as f:
    for line in f:
        if line.startswith(";"):
            continue

        if "[ atoms ]" in line:
            trigger = True
            continue

        elif "[" in line:
            trigger = False

            resname = line.split(";")[0].strip("[ ]\n")

            if resname not in (
                "angles",
                "exclusions",
                "dihedrals",
                "impropers",
                "bonds",
            ):
                db_res[resname] = {}
                db_charges[resname] = {}

        if trigger and line.strip():
            cols = line.strip().split()
            aname, atype, charge = cols[0], cols[1].lower(), cols[2]
            db_res[resname][aname] = atype
            db_charges[resname][aname] = float(charge)

rules_crg = {}

rules["HIS"] = db_res["HI2"]
rules["ASP"] = db_res["AS4"]
rules["LYS"] = db_res["LY3"]
rules["GLU"] = db_res["GL4"]
rules["TYR"] = db_res["TY2"]
rules["CYS"] = db_res["CYS2"]
rules["ARG"] = db_res["AR4"]

for ref, taus in sizes.items():
    for tau in taus:
        db_res[tau] = db_res[ref]

# rules_crg["NTR"] = db_charges["NT3"]
# rules_crg["CTR"] = db_charges["CT4"]
rules_crg["HIS"] = db_charges["HI2"]
rules_crg["ASP"] = db_charges["AS4"]
rules_crg["LYS"] = db_charges["LY3"]
rules_crg["GLU"] = db_charges["GL4"]
rules_crg["TYR"] = db_charges["TY2"]
rules_crg["CYS"] = db_charges["CYS2"]
rules_crg["ARG"] = db_charges["AR4"]

# c6, c12 = db_lj[water_atype.lower()]
# water_radius = radius_from_lennard(c6, c12, rt=rt, water=True)
water_radius = 1.4241233005024403
new_siz = f"""! aminoacids from 54a7
! created by extend_db.py
! water_atype = {water_atype}; rt = {rt}
atom__res_radius_\n"""
with open(old_siz_db) as f:
    for line in f:
        if line.startswith("!") or "atom__res_radius_" in line:
            continue
        cols = line.strip().split()
        aname, resname = cols[0:2]

        if resname in ignore_list:
            continue

        if resname in db_res and aname in db_res[resname]:
            atype = db_res[resname][aname]
        elif resname in rules and aname in rules[resname]:
            atype = rules[resname][aname]
        else:
            print(aname, resname, "ignored on SIZ")
            continue

        sigma, epsilon = db_lj[atype]

        radius = radius_from_lennard(sigma, epsilon, rt=rt)
        out = f"{aname:<4}  {resname:3} {radius:6.3f}\n"

        new_siz += out

with open(out_siz, "w") as f:
    f.write(new_siz)

new_crg = """! aminoacids from 54a7
! created by extend_db.py
!
atom__resnumbc_charge_\n"""
with open(old_crg_db) as f:
    for line in f:
        if line.startswith("!") or "atom__resnumbc_charge_" in line:
            continue

        cols = line.strip().split()
        aname, resname = cols[0:2]

        if resname in ignore_list:
            continue

        if resname in db_charges and aname in db_charges[resname]:
            charge = db_charges[resname][aname]
            if cols[2] != "X.XXX" and float(cols[2]) != charge:
                print("ERROR CRG", aname, resname, charge, cols)
        elif resname[:2] in ("CT", "NT"):
            charge = float(cols[2])
        elif resname in rules_crg and aname in rules_crg[resname]:
            charge = rules_crg[resname][aname]
        else:
            print(aname, resname, "ignored on CRG")
            continue

        out = f"{aname:<4}  {resname:8} {charge:6.3f}\n"

        new_crg += out

with open(out_crg, "w") as f:
    f.write(new_crg)



import os
from copy import copy

from pypka.config import Config
from pypka.constants import (
    LIPID_RESIDUES,
    NUCLEIC_ACIDS,
    RNA_RESIDUES,
    PDB_RNA_RESIDUES,
    IONS,
    TERMINAL_OFFSET,
    TITRABLETAUTOMERS,
)

from pypka.clean.ffconverter import mainchain_Hs, AMBER_Hs, AMBER_mainchain_Hs

from pdbmender.formats import (
    new_gro_line,
    new_pdb_line,
    read_gro_line,
    read_pdb_line,
    read_pqr_line,
)
from pdbmender.utils import (
    mend_pdb,
    add_tautomers,
    rm_cys_bridges,
    get_his_states,
    identify_cter,
)
import logging

logger = logging.getLogger(__name__)


def inputPDBCheck(filename, sites, clean_pdb):
    """
    Returns: chains_length, chains_res
    """
    if filename[-3:] in ("pdb", "pqr"):
        filetype = "pdb"
    elif filename[-3:] == "gro":
        filetype = "gro"
    else:
        raise Exception("Input file must be either a pdb or a gro.")

    chains_length = {}
    chains_res = {chain: {} for chain in sites.keys()}
    chains = []
    if filetype == "pdb" and not clean_pdb:
        new_gro_header = "CREATED within PyPka\n"
        new_gro_body = ""
    with open(filename) as f:
        last_chain = ""
        chain_length = 0

        nline = 0
        maxnlines = 0
        atom_number = 0
        for line in f:
            nline += 1
            atom_line = False
            if filetype == "pdb":
                if line.startswith("ATOM "):
                    atom_line = True
                    chain_length += 1
                    (aname, anumb, resname, chain, resnumb, x, y, z) = read_pdb_line(
                        line
                    )
                    if chain not in chains:
                        chains.append(chain)
                    atom_number += 1
                    if not clean_pdb:
                        if (
                            len(aname) > 2
                            and aname[1] == "H"
                            and aname[0] in ("1", "2")
                        ):
                            aname = aname[1:] + aname[0]
                        new_gro_body += new_gro_line(
                            anumb, aname, resname, resnumb, x / 10.0, y / 10, z / 10
                        )
                elif line.startswith("CRYST1"):
                    tmp = line.split()[1:4]
                    box = (float(tmp[0]), float(tmp[1]), float(tmp[2]))
                    new_gro_footer = "{0:10.5f}{1:10.5f}{2:10.5f}\n".format(
                        box[0] / 10.0, box[1] / 10.0, box[2] / 10.0
                    )

            elif filetype == "gro":
                if nline > 2 and nline < maxnlines:
                    (aname, anumb, resname, resnumb, x, y, z) = read_gro_line(line)
                    chain = "A"
                    atom_line = True
                elif nline == 2:
                    natoms = int(line.strip())
                    maxnlines = natoms + 3

            if atom_line:
                if chain_length == 1:
                    last_chain = chain

                if chain != last_chain and chain_length != 1:
                    chains_length[last_chain] = chain_length
                    # chains_res[chain] = done[chain]
                    chain_length = 0
                    last_chain = chain

                if (
                    chain in sites
                    and resnumb not in chains_res[chain]
                    and str(resnumb) in sites[chain]
                ):
                    if Config.pypka_params["ffinput"] == "CHARMM":
                        if resname in ("HSD", "HSE", "HSP"):
                            resname = "HIS"
                    chains_res[chain][resnumb] = resname

    # if filetype == 'pdb' and not clean_pdb:
    #    new_gro_header += '{0}\n'.format(atom_number)
    #    with open('TMP.gro', 'w') as f:
    #        f.write(new_gro_header + new_gro_body + new_gro_footer)

    chains_length[last_chain] = chain_length
    # chains_res[chain] = done[chain]

    # tmp_chains_res is an ugly hack so that test cases hold
    # TODO: remove tmp_chains_res and update tests
    tmp_chains_res = {}
    for chain in sites.keys():
        if chain not in chains:
            continue

        tmp_chains_res[chain] = {}
        for site in sites[chain]:
            if site[-1] == "C":
                resnumb = site[:-1]
                tmp_chains_res[chain][resnumb] = "CTR"
            elif site[-1] == "N":
                resnumb = site[:-1]
                tmp_chains_res[chain][resnumb] = "NTR"
        chains_res[chain] = {**tmp_chains_res[chain], **chains_res[chain]}

    skipped_sites = {}
    for chain, resnumbs in sites.items():
        if chain not in chains_res:
            skipped_sites[chain] = resnumbs
            continue
        for resnumb in resnumbs:
            skipped = False
            if resnumb[-1] in "NC":
                termini = "NTR" if resnumb[-1] == "N" else "CTR"
                resnumb = resnumb[:-1]
                if (
                    resnumb not in chains_res[chain]
                    or chains_res[chain][resnumb] != termini
                ):
                    print(
                        "{1} in chain '{0}' not found or not titratable.".format(
                            chain, termini
                        )
                    )
            else:
                resnumb = int(resnumb)
                if resnumb not in chains_res[chain].keys():
                    print(
                        "Residue #{1} in chain '{0}' not found or not titratable.".format(
                            chain, resnumb
                        )
                    )

    nsites = sum([len(res) for res in chains_res.values()])
    if not nsites:
        raise Exception(
            "No titrable residues found. Please check the residue number and chain."
        )

    return chains_length, chains_res


def get_pdb_Hs(pdbname, chains_res):
    with open(pdbname) as f:
        content = f.readlines()
    for line in content:
        if line.startswith("ATOM "):
            (
                aname,
                anumb,
                resname,
                chain,
                resnumb,
                x,
                y,
                z,
                _,
                _,
            ) = read_pqr_line(line)

            if chain not in chains_res:
                continue
            elif (
                chain in chains_res
                and resnumb in chains_res[chain]
                and resname in AMBER_Hs
                and aname in AMBER_Hs[resname]
                and aname not in AMBER_mainchain_Hs
            ):
                continue
            elif aname[0] == "H" and aname not in ("H1", "H2", "H3"):
                if not chain in mainchain_Hs:
                    mainchain_Hs[chain] = {}
                if not resnumb in mainchain_Hs[chain]:
                    mainchain_Hs[chain][resnumb] = []
                mainchain_Hs[chain][resnumb].append(
                    (aname, anumb, resname, chain, x, y, z)
                )
    return mainchain_Hs


def cleanPDB(
    pdb_filename,
    final_pdb,
    molecules,
    chains_res,
    automatic_sites,
    inputpqr="clean.pqr",
    outputpqr="cleaned_tau.pqr",
):

    remove_membrane_n_rna(pdb_filename, Config.pypka_params["pdb2pqr_inputfile"])
    if " " in molecules.keys():
        molecules["_"] = molecules[" "]
        del molecules[" "]
        chains_res["_"] = chains_res[" "]
        del chains_res[" "]

    logfile = "LOG_pdb2pqr"

    _ = mend_pdb(
        Config.pypka_params["pdb2pqr_inputfile"],
        inputpqr,
        Config.pypka_params["ff_family"],
        Config.pypka_params["ff_family"],
        logfile=logfile,
        hopt=Config.pypka_params["pdb2pqr_h_opt"],
    )

    chains_res, cys_bridges = rm_cys_bridges(chains_res, logfile)
    for chain in cys_bridges.keys():
        molecule = molecules[chain]
        molecule.saveCYSBridges(cys_bridges[chain])

    # his_states = get_his_states(logfile)
    # molecule.saveHISStates(his_states)

    if automatic_sites:
        old_ctrs = {}
        for chain in chains_res:
            for resnumb, resname in chains_res[chain].items():
                old_ctrs[chain] = []
                if isinstance(resnumb, str) and resname == "CTR":
                    old_ctrs[chain] += [int(resnumb)]

        new_ctrs = identify_cter(inputpqr, old_ctrs)

        for chain, resnumbs in new_ctrs.items():
            for resnumb in resnumbs:
                chains_res[chain][str(resnumb)] = "CTR"
                molecules[chain].CTR += [resnumb]
                sID = molecules[chain].addSite(resnumb + TERMINAL_OFFSET)
                molecules[chain].addTautomers(sID, TITRABLETAUTOMERS["CTR"], "CTR")

    if Config.pypka_params["f_structure_out"]:
        hs_pdb = "Hs.pqr"
        ff_out = Config.pypka_params["ff_structure_out"]
        if ff_out == "gromos_cph":
            ff_out = "gromos"
        mend_pdb(
            Config.pypka_params["pdb2pqr_inputfile"],
            hs_pdb,
            ff_out,
            ff_out,
            logfile=logfile,
        )
        get_pdb_Hs(hs_pdb, chains_res)

    to_exclude = NUCLEIC_ACIDS
    nontitrating_lines = add_tautomers(
        inputpqr,
        chains_res,
        Config.pypka_params["ff_family"],
        outputpqr,
        to_exclude=to_exclude,
        terminal_offset=TERMINAL_OFFSET,
    )

    with open(outputpqr) as f:
        content = f.read()
    with open(outputpqr, "w") as f_new:
        f_new.write(content + nontitrating_lines)

    rna_pqr = None
    write_final_pdb(pdb_filename, outputpqr, final_pdb, rna_pqr)

    tmpfiles = (
        "addhtaut_cleaned.pdb",
        "input_clean_fixed.pdb",
        "LOG_pdb2pqr",
        "clean.pqr",
        "cleaned.pqr",
        "cleaned_tau.pqr",
        "removed.pqr",
        "Hs.pqr",
    )
    for filename in tmpfiles:
        if not Config.debug and os.path.isfile(filename):
            os.remove(filename)
    if (
        not Config.debug
        and os.path.isfile(Config.pypka_params["pdb2pqr_inputfile"])
        and not Config.pypka_params.structure_output
    ):
        os.remove(Config.pypka_params["pdb2pqr_inputfile"])

    return chains_res


def remove_membrane_n_rna(pdbfile, outfile):
    protein_lines = ""
    to_remove = LIPID_RESIDUES + list(Config.pypka_params.LIPIDS.values())

    with open(pdbfile) as f:
        for line in f:
            if line.startswith("ATOM"):
                (aname, anumb, resname, chain, resnumb, x, y, z) = read_pdb_line(line)

                insertion_code = line[26].strip()
                if insertion_code:
                    continue
                if aname[0] == "H" and Config.pypka_params["remove_hs"]:
                    continue

                if Config.pypka_params["ffinput"] == "CHARMM":
                    if resname in ("HSD", "HSE"):
                        resname = "HSP"

                if resname not in to_remove:
                    if resname in PDB_RNA_RESIDUES:
                        resname = PDB_RNA_RESIDUES[resname]
                    protein_lines += new_pdb_line(
                        anumb, aname, resname, resnumb, x, y, z, chain=chain
                    )

            elif line.startswith("ENDMDL"):
                break

    with open(outfile, "w") as f_new:
        f_new.write(protein_lines)

    # rna_fname = None
    # if rna_lines:
    #    rna_fname = "tmp_rna.pdb"
    #    with open(rna_fname, "w") as f_new:
    #        f_new.write(rna_lines)

    return  # rna_fname


def pqr2pdb(line, counter):
    counter += 1
    (aname, anumb_old, resname, chain, resnumb, x, y, z) = read_pdb_line(line)
    if resname in RNA_RESIDUES:
        resname = RNA_RESIDUES[resname]
        incorrect_rna_ops = {"OP1": "O1P", "OP2": "O2P"}
        if aname in incorrect_rna_ops:
            aname = incorrect_rna_ops[aname]
    return (
        new_pdb_line(counter, aname, resname, resnumb, x, y, z, chain=chain),
        counter,
    )


def write_final_pdb(pdb_filename, outputpqr, final_pdb, rna_inputpqr):

    with open(pdb_filename) as f:
        box = None
        for line in f:
            if line.startswith("CRYST1"):
                tmp = line.split()[1:4]
                box = (float(tmp[0]), float(tmp[1]), float(tmp[2]))
        if not box:
            box = (1.0, 1.0, 1.0)

    with open(outputpqr) as f, open(final_pdb, "w") as fnew:
        counter = 0
        new_pdb_text = (
            "CRYST1  {:3.3f}  {:3.3f}  {:3.3f}  "
            "60.00  60.00  90.00 P 1           1\n".format(box[0], box[1], box[2])
        )
        for line in f:
            next_pdb_line, counter = pqr2pdb(line, counter)
            new_pdb_text += next_pdb_line

        if rna_inputpqr:
            with open(rna_inputpqr) as f_rna:
                for line in f_rna:
                    if line.startswith("ATOM "):
                        next_pdb_line, counter = pqr2pdb(line, counter)
                        new_pdb_text += next_pdb_line
        fnew.write(new_pdb_text)

    keep_membrane = False
    keep_ions = False
    if Config.delphi_params["pbc_dim"] == 2:
        keep_membrane = True

    if Config.pypka_params["keep_ions"]:
        keep_ions = True
    if keep_ions or keep_membrane:
        add_non_protein(
            pdb_filename, final_pdb, keep_membrane=keep_membrane, keep_ions=keep_ions
        )

    add_all_deleted(pdb_filename, final_pdb)


def add_all_deleted(pdb_filename, final_pdb):
    def get_all_resnames(pdb_f):
        all_res_types = set()
        with open(pdb_f) as f:
            for line in f:
                if line.startswith("ATOM"):
                    resname = line[17:21].strip()
                    all_res_types.add(resname)
        return all_res_types

    initial_res = get_all_resnames(pdb_filename)
    final_res = get_all_resnames(final_pdb)

    delphi_res = set()
    with open(Config.pypka_params["f_siz"]) as f:
        for line in f:
            if len(line.split()) == 3:
                resname = line[6:10].strip()
                delphi_res.add(resname)

    deleted_res = initial_res - final_res
    re_insert_res = deleted_res.intersection(delphi_res) - set(["CYS"])
    deleted_res = deleted_res - re_insert_res - set(["HISD", "HISH", "HISE", "CYS"])

    for res_type in deleted_res:
        logger.warning("All {} residues have been deleted".format(res_type))

    to_add = []
    with open(pdb_filename) as f:
        for line in f:
            if line.startswith("ATOM"):
                resname = line[17:21].strip()

                if resname in re_insert_res:
                    to_add.append(line)

    with open(final_pdb, "a") as f:
        f.write("".join(to_add))


def add_non_protein(pdbfile_origin, add_to_pdb, keep_membrane=False, keep_ions=False):
    new_file_body = ""

    with open(add_to_pdb) as f:
        for line in f:
            if line.startswith("ATOM "):
                (aname, anumb, resname, chain, resnumb, x, y, z) = read_pdb_line(line)
                last_anumb = anumb
                last_resnumb = resnumb

    # Read the original pdb with the membrane
    with open(pdbfile_origin) as f:
        for line in f:
            if "ATOM " == line[0:5]:
                (aname, anumb, resname, chain, resnumb, x, y, z) = read_pdb_line(line)
                if keep_membrane:
                    if resname in LIPID_RESIDUES:
                        last_anumb += 1
                        new_file_body += new_pdb_line(
                            last_anumb, aname, resname, resnumb, x, y, z, chain=" "
                        )

                    if resname in list(Config.pypka_params.LIPIDS.values()):
                        aname, resname, to_include = convert_FF_atomnames(
                            aname, resname
                        )
                        if to_include:
                            last_anumb += 1
                            resnumb += last_resnumb
                            new_file_body += new_pdb_line(
                                last_anumb,
                                aname,
                                resname,
                                resnumb,
                                x,
                                y,
                                z,
                                chain=" ",
                            )
                if keep_ions and aname in IONS and resname == aname:
                    last_anumb += 1
                    resnumb += last_resnumb
                    new_file_body += new_pdb_line(
                        last_anumb, aname, resname, resnumb, x, y, z, chain=chain
                    )

    with open(add_to_pdb, "a") as f_new:
        f_new.write(new_file_body)


def convert_FF_atomnames(aname, resname):
    popc_resname = Config.pypka_params.LIPIDS["POPC"]
    chol_resname = Config.pypka_params.LIPIDS["cholesterol"]

    lookup = [
        ((" N  ", popc_resname), ("NTM", "CHL")),
        ((" C12", popc_resname), ("CB ", "CHL")),
        ((" C13", popc_resname), ("CN1", "CHL")),
        ((" C14", popc_resname), ("CN2", "CHL")),
        ((" C15", popc_resname), ("CN3", "CHL")),
        ((" C11", popc_resname), ("CA ", "PJ2")),
        ((" P  ", popc_resname), ("P  ", "PJ2")),
        ((" O13", popc_resname), ("OG ", "PJ2")),
        ((" O14", popc_resname), ("OB ", "PJ2")),
        ((" O12", popc_resname), ("OA ", "PJ2")),
        ((" O11", popc_resname), ("OD ", "PJ2")),
        ((" C1 ", popc_resname), ("CD ", "POX")),
        ((" C2 ", popc_resname), ("CE ", "POX")),
        ((" O21", popc_resname), ("OE ", "POX")),
        ((" C21", popc_resname), ("C1A", "POX")),
        ((" O22", popc_resname), ("O1A", "POX")),
        ((" C22", popc_resname), ("C1B", "POX")),
        ((" C3 ", popc_resname), ("CZ ", "POX")),
        ((" O31", popc_resname), ("OZ ", "POX")),
        ((" C31", popc_resname), ("C2A", "POX")),
        ((" O32", popc_resname), ("O2A", "POX")),
        ((" C32", popc_resname), ("C2B", "POX")),
        ((" C23", popc_resname), ("C1C", "POX")),
        ((" C24", popc_resname), ("C1D", "POX")),
        ((" C25", popc_resname), ("C1E", "POX")),
        ((" C26", popc_resname), ("C1F", "POX")),
        ((" C27", popc_resname), ("C1G", "POX")),
        ((" C28", popc_resname), ("C1H", "POX")),
        ((" C29", popc_resname), ("C1I", "POX")),
        (("0C21", popc_resname), ("C1J", "POX")),
        (("1C21", popc_resname), ("C1K", "POX")),
        (("2C21", popc_resname), ("C1L", "POX")),
        (("3C21", popc_resname), ("C1M", "POX")),
        (("4C21", popc_resname), ("C1N", "POX")),
        (("5C21", popc_resname), ("C1O", "POX")),
        (("6C21", popc_resname), ("C1P", "POX")),
        (("7C21", popc_resname), ("C1Q", "POX")),
        (("8C21", popc_resname), ("C1R", "POX")),
        ((" C33", popc_resname), ("C2C", "POX")),
        ((" C34", popc_resname), ("C2D", "POX")),
        ((" C35", popc_resname), ("C2E", "POX")),
        ((" C36", popc_resname), ("C2F", "POX")),
        ((" C37", popc_resname), ("C2G", "POX")),
        ((" C38", popc_resname), ("C2H", "POX")),
        ((" C39", popc_resname), ("C2I", "POX")),
        (("0C31", popc_resname), ("C2J", "POX")),
        (("1C31", popc_resname), ("C2K", "POX")),
        (("2C31", popc_resname), ("C2L", "POX")),
        (("3C31", popc_resname), ("C2M", "POX")),
        (("4C31", popc_resname), ("C2N", "POX")),
        (("5C31", popc_resname), ("C2O", "POX")),
        (("6C31", popc_resname), ("C2P", "POX")),
        ((" H3", chol_resname), (" H3", "CHO")),
        ((" C3", chol_resname), (" C3", "CHO")),
        ((" C2", chol_resname), (" C2", "CHO")),
        ((" C1", chol_resname), (" C1", "CHO")),
        (("C10", chol_resname), ("C10", "CHO")),
        (("C19", chol_resname), ("C19", "CHO")),
        ((" C9", chol_resname), (" C9", "CHO")),
        (("C11", chol_resname), ("C11", "CHO")),
        (("C12", chol_resname), ("C12", "CHO")),
        (("C13", chol_resname), ("C13", "CHO")),
        (("C18", chol_resname), ("C18", "CHO")),
        (("C17", chol_resname), ("C17", "CHO")),
        (("C20", chol_resname), ("C20", "CHO")),
        (("C22", chol_resname), ("C22", "CHO")),
        (("C23", chol_resname), ("C23", "CHO")),
        (("C24", chol_resname), ("C24", "CHO")),
        (("C25", chol_resname), ("C25", "CHO")),
        (("C27", chol_resname), ("C27", "CHO")),
        (("C26", chol_resname), ("C26", "CHO")),
        (("C21", chol_resname), ("C21", "CHO")),
        (("C16", chol_resname), ("C16", "CHO")),
        (("C15", chol_resname), ("C15", "CHO")),
        (("C14", chol_resname), ("C14", "CHO")),
        ((" C8", chol_resname), (" C8", "CHO")),
        ((" C7", chol_resname), (" C7", "CHO")),
        ((" C6", chol_resname), (" C6", "CHO")),
        ((" H6", chol_resname), (" H6", "CHO")),
        ((" C5", chol_resname), (" C5", "CHO")),
        ((" C4", chol_resname), (" C4", "CHO")),
        ((" O3", chol_resname), (" O3", "CHO")),
        (("H3'", chol_resname), ("H3'", "CHO")),
    ]

    for swap in lookup:
        old_aname = swap[0][0].strip()
        old_resname = swap[0][1]
        new_aname = swap[1][0].strip()
        new_resname = swap[1][1]
        if old_aname == aname and old_resname == resname:
            aname = new_aname
            resname = new_resname
            return aname, resname, True

    return aname, resname, False



import os
import logging
from pypka.clean.ffconverter import (
    AMBER_protomers,
    GROMOS_protomers,
    gromos2amber,
    mainchain_Hs,
)
from pypka.config import Config
from pypka.constants import *
from pdbmender.formats import new_pdb_line, read_pdb_line, read_pqr_line
from pdbmender.utils import mend_pdb

logger = logging.getLogger(__name__)


def write_output_structure(sites, molecules, delphi_input_content):
    def getProtomerResname(pdb_content, site, pH, ff_protomers):
        resnumb = site.getResNumber()
        resname = site.getName()
        new_state, (state_prob, taut_prob) = site.getMostProbTaut(pH)
        new_state_i = new_state - 1
        for ff_resname, protomers in ff_protomers[resname].items():
            if new_state_i in protomers.keys():
                new_resname = ff_resname
                remove_hs = protomers[new_state_i]
                average_prot = site.getTitrationCurve()[pH]
                if state_prob < 0.75:
                    warn = (
                        "{0}{1} "
                        "protonation state probability: {2}, "
                        "tautomer probability: {3}".format(
                            resname, resnumb, state_prob, taut_prob
                        )
                    )
                    logger.warning(warn)

                rounded_sprob = round(state_prob, 2)
                rounded_tprob = round(taut_prob, 2)
                rounded_avgprot = round(average_prot, 2)
                remark_line = (
                    "{0: <5}{1: <6}    {2: >1.2f}         "
                    "{3: >1.2f}         {4: >1.2f}".format(
                        resname,
                        resnumb,
                        rounded_avgprot,
                        rounded_sprob,
                        rounded_tprob,
                    )
                )
                pdb_content += "REMARK     {text}\n".format(text=remark_line)
        # print(resnumb, new_state, new_resname, remove_hs, state_prob, taut_prob)
        return pdb_content, new_state_i, new_resname, remove_hs

    outputname = Config.pypka_params["f_structure_out"]
    pH = float(Config.pypka_params["f_structure_out_pH"])
    ff_out = Config.pypka_params["ff_structure_out"]
    ff_protomer = {"amber": AMBER_protomers, "gromos_cph": GROMOS_protomers}[ff_out]
    pdb_content = (
        f"REMARK     PypKa assigned protonation states @ pH {pH}\n"
        "REMARK     Residue    Avg Prot   State Prob    Taut Prob\n"
    )
    new_states = {}
    for site in sites:
        resname = site.getName()
        resnumb = site.res_number
        molecule = site.molecule
        chain = molecule.chain
        (pdb_content, new_state, new_resname, remove_hs) = getProtomerResname(
            pdb_content, site, pH, ff_protomer
        )
        if resname in ("NTR", "CTR"):
            new_resname = site.termini_resname
        if chain not in new_states:
            new_states[chain] = {}
        new_states[chain][resnumb] = (resname, new_state, new_resname, remove_hs)
    new_pdb = pdb_content
    counter = 0
    tit_atoms = {}
    other_atoms = {}
    for molecule in molecules.values():
        for atom_numb in molecule.atoms_tit_res:
            if molecule.atoms_tit_res[atom_numb]:
                tit_atoms[atom_numb] = molecule
            else:
                other_atoms[atom_numb] = molecule

    in_delphi_pdb = {}
    for line in delphi_input_content:
        if line.startswith("ATOM "):
            (aname, anumb, resname, chain, resnumb, x, y, z) = read_pdb_line(line)
            if chain not in in_delphi_pdb:
                in_delphi_pdb[chain] = {}
            if resnumb not in in_delphi_pdb[chain]:
                in_delphi_pdb[chain][resnumb] = []
            in_delphi_pdb[chain][resnumb].append(aname)

    for line in delphi_input_content:
        if line.startswith("ATOM "):
            (aname, anumb, resname, chain, resnumb, x, y, z) = read_pdb_line(line)

            if anumb in tit_atoms.keys():
                molecule = tit_atoms[anumb]
                (oldresname, new_state, resname, removeHs) = new_states[chain][resnumb]
                if aname in removeHs:
                    continue
                if (
                    ff_out == "amber"
                    and oldresname in gromos2amber
                    and new_state in gromos2amber[oldresname]
                    and aname in gromos2amber[oldresname][new_state]
                ):
                    aname = gromos2amber[oldresname][new_state][aname]
            elif anumb in other_atoms:
                molecule = other_atoms[anumb]
            else:
                continue
            if resnumb > TERMINAL_OFFSET:
                termini_site = molecule.sites[resnumb]
                resnumb -= TERMINAL_OFFSET
                if resnumb in molecule.sites.keys():
                    _, ter_new_state, resname, ter_removeHs = new_states[chain][resnumb]
                else:
                    resname = termini_site.termini_resname
                # print(new_pdb_line(anumb, aname, resname, resnumb, x, y, z).strip())
            if resnumb in molecule.getCYS_bridges():
                resname = "CYX"
            counter += 1
            new_pdb += new_pdb_line(
                counter, aname, resname, resnumb, x, y, z, chain=chain
            )
            if chain in mainchain_Hs and resnumb in mainchain_Hs[chain]:
                while len(mainchain_Hs[chain][resnumb]) > 0:
                    counter += 1
                    (aname, anumb, oldresname, chain, x, y, z) = mainchain_Hs[chain][
                        resnumb
                    ].pop()
                    if (
                        resnumb not in in_delphi_pdb[chain]
                        or aname not in in_delphi_pdb[chain][resnumb]
                    ):
                        new_pdb += new_pdb_line(
                            counter, aname, resname, resnumb, x, y, z, chain=chain
                        )
                del mainchain_Hs[chain][resnumb]
        elif not line.startswith("ENDMDL"):
            new_pdb += line

    outputpqr = "leftovers.pqr"
    logfile = "LOG_pdb2pqr_nontitrating"
    if ff_out == "gromos_cph":
        ff_out = "GROMOS"
    mend_pdb(
        Config.pypka_params["pdb2pqr_inputfile"],
        outputpqr,
        ff_out,
        ff_out,
        logfile=logfile,
    )
    os.system("rm -f input_clean_fixed.pdb")

    with open(outputpqr) as f:
        for line in f:
            if line.startswith("ATOM "):
                (
                    aname,
                    anumb,
                    resname,
                    chain,
                    resnumb,
                    x,
                    y,
                    z,
                    charge,
                    radius,
                ) = read_pqr_line(line)
                if chain not in mainchain_Hs:
                    counter += 1
                    new_pdb += new_pdb_line(
                        counter, aname, resname, resnumb, x, y, z, chain=chain
                    )
    to_remove = (logfile, outputpqr, Config.pypka_params["pdb2pqr_inputfile"])
    for f in to_remove:
        os.remove(f)
    with open(outputname, "w") as f_new:
        f_new.write(new_pdb)



import os
from pypka.config import Config


def checkDelPhiErrors(filename, mode=None):
    exceptions = []
    if mode == "readFiles":
        exceptions = ["part of system outside the box!", "has a net charge of"]
    elif mode == "runDelPhi":
        exceptions = ["part of system outside the box!"]

    exit_trigger = False
    errors = ""
    with open(filename) as f:
        for line in f:
            ignore_error = False
            if "WARNING".lower() in line.lower() or "ERROR".lower() in line.lower():
                for exception in exceptions:
                    if exception in line:
                        ignore_error = True
                if not ignore_error:
                    exit_trigger = True
                    errors += line
    if exit_trigger:
        raise Exception(
            "The following errors have been found on {0}: \n{1}".format(
                filename, errors
            )
        )
    if not Config.debug:
        if os.path.isfile(filename):
            os.remove(filename)
        focusing_log = "{}_focusing".format(filename)
        if os.path.isfile(focusing_log):
            os.remove(focusing_log)



import sys
import os
import pathlib

# sys.path.append(os.path.dirname(os.path.realpath(__file__)) + "../")
sys.path.insert(0, str(pathlib.Path(__file__).parent.parent))

from pypka.cli import CLI

if __name__ == "__main__":
    CLI()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pypka', 'pypka.CHARMM36m', 'pypka.G54A7', 'pypka.clean', 'pypka.mc']

package_data = \
{'': ['*'],
 'pypka.CHARMM36m': ['sts/*', 'sts_cphmd/*'],
 'pypka.G54A7': ['sts/*', 'sts_cphmd/*', 'sts_old/*']}

install_requires = \
['delphi4py==1.2.0', 'numpy', 'pdbmender==0.5.3', 'psutil', 'pytest>=6.2.5']

setup_kwargs = {
    'name': 'pypka',
    'version': '2.9.4',
    'description': 'A python module for flexible Poisson-Boltzmann based pKa calculations with proton tautomerism',
    'long_description': '[![Open In Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/github/mms-fcul/PypKa/blob/master/examples/notebook/pypka.ipynb) [![CircleCI](https://circleci.com/gh/mms-fcul/PypKa.svg?style=svg)](https://circleci.com/gh/mms-fcul/PypKa) [![Codacy Badge](https://app.codacy.com/project/badge/Grade/59a058e4bf0846f18d9d1f6b16a4a0e5)](https://www.codacy.com/gh/mms-fcul/PypKa/dashboard?utm_source=github.com&amp;utm_medium=referral&amp;utm_content=mms-fcul/PypKa&amp;utm_campaign=Badge_Grade) [![Documentation Status](https://readthedocs.org/projects/pypka/badge/?version=latest)](https://pypka.readthedocs.io/en/latest/?badge=latest)\n\n[![PyPI version](https://badge.fury.io/py/pypka.svg)](https://badge.fury.io/py/pypka)  [![PyPI - Downloads](https://img.shields.io/pypi/dm/pypKa)](https://badge.fury.io/py/pypKa)\n\n# PypKa\n\nA python module for flexible Poisson-Boltzmann based pKa calculations with proton tautomerism\nDOI: <a href="https://doi.org/10.1021/acs.jcim.0c00718">10.1021/acs.jcim.0c00718</a>\n\n```bibtex\n@article{reis2020jcim,\nauthor = {Reis, Pedro B. P. S. and Vila-Viçosa, Diogo and Rocchia, Walter and Machuqueiro, Miguel},\ntitle = {PypKa: A Flexible Python Module for Poisson–Boltzmann-Based pKa Calculations},\njournal = {Journal of Chemical Information and Modeling},\nvolume = {60},\nnumber = {10},\npages = {4442-4448},\nyear = {2020},\ndoi = {10.1021/acs.jcim.0c00718}\n}\n```\n\n## Documentation\n\n  Documentation can be found [here](https://pypka.readthedocs.io/en/latest/)\n\n  Starting templates for the the API and CLI usage can be found [here](https://pypka.readthedocs.io/en/latest/example.html) while a online notebook is hosted at [Google Colab](https://colab.research.google.com/github/mms-fcul/PypKa/blob/master/pypka/example/notebook/pypka.ipynb)\n\n\n## Contributing\n\nContributions are encouraged, and they are greatly appreciated!\n\nYou can contribute in many ways:\n\n* Report Bugs\n* Fix Bugs\n* Implement Features\n* Write Documentation\n* Submit Feedback\n\nFor more info check [CONTRIBUTING](./CONTRIBUTING.rst)\n\n## License\n\n  pypka is distributed under a [LGPL-3.0](./LICENSE), however delphi4py depends on\n  DelPhi which is proprietary. To use DelPhi the user is required to\n  download the DelPhi license\n  [here](https://honiglab.c2b2.columbia.edu/software/cgi-bin/software.pl?input=DelPhi)\n\n## Contacts\n\n  Please submit a github issue to report bugs and to request new features.\n  Alternatively you may find the developer [here](mailto:pdreis@fc.ul.pt). Please visit ou [website](http://mms.rd.ciencias.ulisboa.pt/) for more information.\n',
    'author': 'Pedro Reis',
    'author_email': 'pdreis@fc.ul.pt',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://pypka.org',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.2,<4.0',
}


setup(**setup_kwargs)
